"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("underscore");
const mongoose_1 = require("mongoose");
/**
 * Wrapped Model from mongoose.Model.
 */
class Model {
    static Config(config) {
        this.$CONFIG = config;
        return this;
    }
    static Schema(schema) {
        this.$SCHEMA = schema;
        return this;
    }
    static Pre(pre) {
        if (this.hasOwnProperty('$PRES')) {
            this.$PRES.push(pre);
        }
        else {
            this.$PRES = [pre];
        }
        return this;
    }
    static Post(post) {
        if (this.hasOwnProperty('$POSTS')) {
            this.$POSTS.push(post);
        }
        else {
            this.$POSTS = [post];
        }
        return this;
    }
    static Virtual(virtual) {
        if (this.hasOwnProperty('$VIRTUALS')) {
            this.$VIRTUALS.push(virtual);
        }
        else {
            this.$VIRTUALS = [virtual];
        }
        return this;
    }
    static Plugin(plugin) {
        if (this.hasOwnProperty('$PLUGINS')) {
            this.$PLUGINS.push(plugin);
        }
        else {
            this.$PLUGINS = [plugin];
        }
        return this;
    }
    static Index(index) {
        if (this.hasOwnProperty('$INDEXES')) {
            this.$INDEXES.push(index);
        }
        else {
            this.$INDEXES = [index];
        }
        return this;
    }
    static Mixin(model) {
        if (this.hasOwnProperty('$MIXINS')) {
            this.$MIXINS.push(model);
        }
        else {
            this.$MIXINS = [model];
        }
        return this;
    }
    static _$modifySchema(discriminatorKey) {
        if (!this.hasOwnProperty('$SCHEMA')) {
            this.$SCHEMA = {};
        }
        let uniqueFields = [];
        for (let name in this.$SCHEMA) {
            let opt = this.$SCHEMA[name];
            if (opt.unique) {
                opt.unique = false;
                uniqueFields.push(name);
            }
        }
        for (let field of uniqueFields) {
            let fieldUniqueKey = `${field}_unique`;
            let fields = {};
            fields[discriminatorKey] = 1;
            fields[field] = 1;
            fields[fieldUniqueKey] = '2dsphere';
            this.$SCHEMA[fieldUniqueKey] = {
                type: PointSchema,
                // api:        AUTOGEN,
                default: PointSchema,
            };
            this.Index({
                fields,
                options: {
                    unique: true
                }
            });
        }
        this.$SCHEMA[discriminatorKey] = { type: String };
        this.Pre({
            name: 'save',
            fn: setDiscriminatorKey,
        });
        for (let name of exports.preQueries) {
            this.Pre({ name, fn: patchDiscriminatorKey });
        }
        function setDiscriminatorKey(next) {
            if (this[discriminatorKey] == null) {
                this[discriminatorKey] = this.constructor.modelName;
            }
            next();
        }
        function patchDiscriminatorKey() {
            this._conditions[discriminatorKey] = this.model.modelName;
        }
    }
    static $mongooseOptions() {
        this._$initialize();
        if (this._mongooseOptions.initialized || this === Model) {
            return this._mongooseOptions;
        }
        let superClass = this.__proto__;
        let superOptions = (superClass.$mongooseOptions ? superClass.$mongooseOptions() : {
            initialized: false
        });
        if (superOptions.config.discriminatorKey) {
            this._$modifySchema(superOptions.config.discriminatorKey);
        }
        let mixinModels = this.hasOwnProperty('$MIXINS') ? this.$MIXINS : [];
        let mixinOptions = _.map(mixinModels, (model) => model.$mongooseOptions());
        let mixinSchemas = _.map(mixinOptions, (opt) => opt.schema);
        let flattenSchemas = _.flatten([{}, mixinSchemas, superOptions.schema, this.$SCHEMA]);
        this._mongooseOptions.schema = _.extend.apply(null, flattenSchemas);
        let mixinConfigs = _.map(mixinOptions, (opt) => opt.config);
        let flattenConfigs = _.flatten([{}, mixinConfigs, superOptions.config, this.$CONFIG]);
        this._mongooseOptions.config = _.extend.apply(null, flattenConfigs);
        let mixinPres = _.map(mixinOptions, (opt) => opt.pres);
        this._mongooseOptions.pres = _.union(_.flatten([
            mixinPres, superOptions.pres || [],
            this.hasOwnProperty('$PRES') ? this.$PRES : []
        ]));
        let mixinPosts = _.map(mixinOptions, (opt) => opt.posts);
        this._mongooseOptions.posts = _.union(_.flatten([
            mixinPosts, superOptions.posts || [],
            this.hasOwnProperty('$POSTS') ? this.$POSTS : []
        ]));
        let mixinVirtuals = _.map(mixinOptions, (opt) => opt.virtuals);
        this._mongooseOptions.virtuals = _.union(_.flatten([
            mixinVirtuals, superOptions.virtuals || [],
            this.hasOwnProperty('$VIRTUALS') ? this.$VIRTUALS : []
        ]));
        let mixinPlugins = _.map(mixinOptions, (opt) => opt.plugins);
        this._mongooseOptions.plugins = _.union(_.flatten([
            mixinPlugins, superOptions.plugins || [],
            this.hasOwnProperty('$PLUGINS') ? this.$PLUGINS : []
        ]));
        let mixinIndexes = _.map(mixinOptions, (opt) => opt.indexes);
        this._mongooseOptions.indexes = _.union(_.flatten([
            mixinIndexes, superOptions.indexes || [],
            this.hasOwnProperty('$INDEXES') ? this.$INDEXES : []
        ]));
        let mixinMethods = _.map(mixinOptions, (opt) => opt.methods);
        this._mongooseOptions.methods = _.union(_.flatten([
            mixinMethods, superOptions.methods || []
        ]));
        let mixinStatics = _.map(mixinOptions, (opt) => opt.statics);
        this._mongooseOptions.statics = _.union(_.flatten([
            mixinStatics, superOptions.statics || []
        ]));
        for (let name of Object.getOwnPropertyNames(this.prototype)) {
            if (name === 'constructor') {
                continue;
            }
            let descriptor = Object.getOwnPropertyDescriptor(this.prototype, name);
            if (descriptor.value && _.isFunction(descriptor.value)) {
                this._mongooseOptions.methods.push({
                    name: name,
                    fn: descriptor.value,
                });
            }
            if (descriptor.get || descriptor.set) {
                this._mongooseOptions.virtuals.push({
                    name: name,
                    get: descriptor.get,
                    set: descriptor.set,
                });
            }
        }
        for (let name of Object.getOwnPropertyNames(this)) {
            let descriptor = Object.getOwnPropertyDescriptor(this, name);
            if (descriptor.value && _.isFunction(descriptor.value)) {
                this._mongooseOptions.statics.push({
                    name: name,
                    fn: descriptor.value,
                });
            }
        }
        var mongooseSchema = new mongoose_1.Schema(this._mongooseOptions.schema, _.clone(this._mongooseOptions.config));
        mongooseSchema.parentSchema = superOptions.mongooseSchema;
        for (let pre of this._mongooseOptions.pres) {
            pre.parallel = pre.parallel || false;
            mongooseSchema.pre(pre.name, pre.parallel, pre.fn, pre.errorCb);
        }
        for (let post of this._mongooseOptions.posts) {
            mongooseSchema.post(post.name, post.fn);
        }
        for (let virtual of this._mongooseOptions.virtuals) {
            let v = mongooseSchema.virtual(virtual.name);
            if (virtual.get) {
                v = v.get(virtual.get);
            }
            if (virtual.set) {
                v = v.set(virtual.set);
            }
        }
        for (let method of this._mongooseOptions.methods) {
            mongooseSchema.methods[method.name] = method.fn;
        }
        for (let method of this._mongooseOptions.statics) {
            mongooseSchema.statics[method.name] = method.fn;
        }
        for (let plugin of this._mongooseOptions.plugins) {
            let options = _.extend({}, this._mongooseOptions.config, plugin.options);
            mongooseSchema.plugin(plugin.fn, options);
        }
        for (let index of this._mongooseOptions.indexes) {
            mongooseSchema.index(index);
        }
        this._mongooseOptions.mongooseSchema = mongooseSchema;
        this._mongooseOptions.initialized = true;
        return this._mongooseOptions;
    }
    static $register(mongooseInstance) {
        if (!mongooseInstance) {
            mongooseInstance = require('mongoose');
        }
        let model = mongooseInstance.model(this.name, this.$mongooseOptions().mongooseSchema);
        return model;
    }
    static _$initialize() {
        if (this.hasOwnProperty('_mongooseOptions')) {
            return;
        }
        this._mongooseOptions = {
            initialized: false,
            config: {},
        };
    }
    static cast() {
        return this;
    }
}
exports.Model = Model;
class NativeError extends global.Error {
}
exports.NativeError = NativeError;
exports.preQueries = [
    'find', 'findOne', 'count', 'findOneAndUpdate', 'findOneAndRemove', 'update'
];
const PointSchema = new mongoose_1.Schema({
    type: {
        type: String,
        enum: ['Point'],
        default: 'Point',
    },
    coordinates: {
        type: [Number],
        default: [0, 0],
    }
});
// Patch MModel.prototype.init
let _init = mongoose_1.Model.prototype.init;
mongoose_1.Model.cast = Model.cast;
mongoose_1.Model.prototype.init = function (doc, query, fn) {
    let discriminatorKey = this.schema.options.discriminatorKey;
    let type = doc[discriminatorKey];
    if (type) {
        let model = this.db.model(type);
        if (model) {
            this.schema = model.schema;
            this.__proto__ = model.prototype;
            _init.call(this, doc, query);
            if (fn) {
                fn(null);
            }
            return this;
        }
    }
    _init.call(this, doc, query, fn);
};

//# sourceMappingURL=model.js.map
